#include "stdafx.h"

#include "FFChoosePixelFormat.h"


#if WIN32
	#include <Windows.h>
	#include <gl/gl.h>
	#include <gl/wglext.h>
#else
	#include <OpenGL/gl.h>
	#include <OpenGL/glu.h>
	#include <AGL/agl.h>
#endif

#if QUESA_OS_COCOA
	#include "CocoaChoosePixelFormat.h"
#endif

#include "StDummyWindowGLContext.h"

#include <string>
#include <cstring>
#include <sstream>

#ifndef ARRAY_SIZE
	#define		ARRAY_SIZE(x)	(sizeof(x)/sizeof(x[0]))
#endif

namespace
{
	struct PFInfo
	{
		bool	isAccelerated;
		bool	hasStencil;
		bool	hasStereo;
		int		samples;
		int		depthBits;
		int		colorBits;
	};
}
#if WIN32
#pragma mark -

static PFNWGLGETPIXELFORMATATTRIBIVARBPROC	wglGetPixelFormatAttribivARB = NULL;

static PFNWGLCHOOSEPIXELFORMATARBPROC	wglChoosePixelFormatARB = NULL;


static bool IsFormatBetter( const PFInfo& inNew, const PFInfo& inOld,
							SBPixelFormatOptions inOptions )
{
	bool	isBetter = false;
	
	if (inNew.isAccelerated != inOld.isAccelerated)
	{
		if ( (inOptions & kSBContextFormat_PreferAccelerated) != 0 )
		{
			isBetter = inNew.isAccelerated;
		}
		else
		{
			isBetter = not inNew.isAccelerated;
		}
		return isBetter;
	}
	
	if ( (inOptions & kSBContextFormat_PreferStencil) != 0 )
	{
		if (inNew.hasStencil != inOld.hasStencil)
		{
			isBetter = inNew.hasStencil;
			return isBetter;
		}
	}
	
	if ( (inOptions & kSBContextFormat_Stereo) != 0 )
	{
		if (inNew.hasStereo != inOld.hasStereo)
		{
			isBetter = inNew.hasStereo;
			return isBetter;
		}
	}
	else	// do not want quad buffered stereo
	{
		if (inNew.hasStereo != inOld.hasStereo)
		{
			isBetter = not inNew.hasStereo;
			return isBetter;
		}
	}
	
	if ( (inOptions & kSBContextFormat_PreferMultisample) != 0 )
	{
		// 2 or more samples is better than 1...
		bool	isNew1Sample = (inNew.samples <= 1);
		bool	isOld1Sample = (inOld.samples <= 1);
		
		if (isNew1Sample != isOld1Sample)
		{
			isBetter = not isNew1Sample;
			return isBetter;
		}
		
		// but otherwise fewer is better (for speed).
		if (inNew.samples != inOld.samples)
		{
			return (inNew.samples < inOld.samples);
		}
	}
	
	if (inNew.depthBits != inOld.depthBits)
	{
		return (inNew.depthBits > inOld.depthBits);
	}
	
	return (inNew.colorBits > inOld.colorBits);
}

static bool WinSupportsMultisample()
{
	bool	mayMultisample = false;

	// Get the major and minor parts of the OpenGL version number.  We won't
	// need the 3rd part, if any.
	float	theGLVersion = 0.0f;
	const char* versStr = (const char*) glGetString( GL_VERSION );
	if (versStr != NULL)
	{
		std::istringstream	iss( versStr );
		iss >> theGLVersion;
	}
	
	if (theGLVersion >= 1.3f)
	{
		mayMultisample = true;
	}
	else
	{
		// Get the extension string, padded with blanks.
		const char*	extStr = (const char*) glGetString( GL_EXTENSIONS );
		std::string	theExtensions( " " );
		if (extStr != NULL)
		{
			theExtensions += extStr;
			theExtensions += ' ';
		}
		
		if (theExtensions.find(" GL_ARB_multisample ") != std::string::npos)
		{
			mayMultisample = true;
		}
	}
	
	return mayMultisample;
}

enum EAttIndex
{
	kAttIndex_window = 0,
	kAttIndex_bitmap,
	kAttIndex_acceleration,
	kAttIndex_opengl,
	kAttIndex_double_buffer,
	kAttIndex_pixel_type,
	kAttIndex_color_bits,
	kAttIndex_depth_bits,
	kAttIndex_stencil_bits,
	kAttIndex_stereo,
	kAttIndex_samples,
	kAttIndex_sample_buffers
};

static FFPixelFormat_Win	WinChooseFormatNew( HDC inDC, SBPixelFormatOptions inOptions )
{
	FFPixelFormat_Win	bestFormat = 0;
	PFInfo	bestInfo, curInfo;
	
	int	numFormats = 0;
	int	askCountAtt = WGL_NUMBER_PIXEL_FORMATS_ARB;
	(*wglGetPixelFormatAttribivARB)( inDC, 0, 0, 1, &askCountAtt, &numFormats );
	bool	testMultisample =
		((inOptions & kSBContextFormat_PreferMultisample) != 0) and
		WinSupportsMultisample();
	
	int	requestAtts[] =
	{
		WGL_DRAW_TO_WINDOW_ARB,
		WGL_DRAW_TO_BITMAP_ARB,
		WGL_ACCELERATION_ARB,
		WGL_SUPPORT_OPENGL_ARB,
		WGL_DOUBLE_BUFFER_ARB,
		WGL_PIXEL_TYPE_ARB,
		WGL_COLOR_BITS_ARB,
		WGL_DEPTH_BITS_ARB,
		WGL_STENCIL_BITS_ARB,
		WGL_STEREO_ARB,
		WGL_SAMPLES_ARB,			// keep these 2 last
		WGL_SAMPLE_BUFFERS_ARB
	};
	
	int	numAtts = ARRAY_SIZE( requestAtts );
	if (not testMultisample)
	{
		numAtts -= 2;
	}
	int	attValues[ ARRAY_SIZE( requestAtts ) ];
	
	for (FFPixelFormat_Win fmt = 1; fmt <= numFormats; ++fmt)
	{
		if ((*wglGetPixelFormatAttribivARB)( inDC, fmt, 0, numAtts,
			requestAtts, attValues ))
		{
			if (not attValues[ kAttIndex_opengl ])
			{
				continue;
			}
			
			if (attValues[ kAttIndex_pixel_type ] != WGL_TYPE_RGBA_ARB)
			{
				continue;
			}
			
			if (attValues[ kAttIndex_depth_bits ] < 16)
			{
				continue;
			}
			
			if ( (inOptions & kSBContextFormat_Window) != 0 )
			{
				if (not attValues[ kAttIndex_window ])
				{
					continue;
				}
				
				if ( (inOptions & kSBContextFormat_SingleBuffer) != 0 )
				{
					if (attValues[ kAttIndex_double_buffer ])
					{
						continue;
					}
				}
				else	// want double buffer
				{
					if (not attValues[ kAttIndex_double_buffer ])
					{
						continue;
					}
				}
			}
			else
			{
				if (not attValues[ kAttIndex_bitmap ])
				{
					continue;
				}
			}
			
			curInfo.isAccelerated = (attValues[ kAttIndex_acceleration ] ==
				WGL_FULL_ACCELERATION_ARB );
			curInfo.hasStencil = (attValues[ kAttIndex_stencil_bits ] >= 8);
			curInfo.hasStereo = (attValues[ kAttIndex_stereo ] > 0);
			curInfo.samples = testMultisample and
				(attValues[ kAttIndex_sample_buffers ] > 0)?
				attValues[ kAttIndex_samples ] : 0;
			curInfo.depthBits = attValues[ kAttIndex_depth_bits ];
			curInfo.colorBits = attValues[ kAttIndex_color_bits ];
			
			if ( (bestFormat == 0) or
				IsFormatBetter( curInfo, bestInfo, inOptions ) )
			{
				bestFormat = fmt;
				bestInfo = curInfo;
			}
		}
	}
	
	return bestFormat;
}

static FFPixelFormat_Win	WinChooseFormatOld( HDC inDC, SBPixelFormatOptions inOptions )
{
	FFPixelFormat_Win	bestFormat = 0;
	PFInfo	bestInfo, curInfo;
	
	int	numFormats = DescribePixelFormat( inDC, 1, sizeof(PIXELFORMATDESCRIPTOR), NULL );
	
	PIXELFORMATDESCRIPTOR	pixelFormatDesc;
	
	for (FFPixelFormat_Win fmt = 1; fmt <= numFormats; ++fmt)
	{
		std::memset( &pixelFormatDesc, 0, sizeof(pixelFormatDesc) );
		pixelFormatDesc.nSize      = sizeof(pixelFormatDesc);
		pixelFormatDesc.nVersion   = 1;
		
		DescribePixelFormat( inDC, fmt, sizeof(PIXELFORMATDESCRIPTOR), &pixelFormatDesc );
		
		if ( (pixelFormatDesc.dwFlags & PFD_SUPPORT_OPENGL) == 0 )
		{
			continue;
		}
		
		if ( pixelFormatDesc.iPixelType != PFD_TYPE_RGBA )
		{
			continue;
		}
		
		if ( pixelFormatDesc.cDepthBits < 16 )
		{
			continue;
		}
		
		if ( (inOptions & kSBContextFormat_Window) != 0 )
		{
			if ( (pixelFormatDesc.dwFlags & PFD_DRAW_TO_WINDOW) == 0 )
			{
				continue;
			}
			if ( (pixelFormatDesc.dwFlags & PFD_DOUBLEBUFFER) == 0 )
			{
				continue;
			}
		}
		else
		{
			if ( (pixelFormatDesc.dwFlags & PFD_DRAW_TO_BITMAP) == 0 )
			{
				continue;
			}
		}
		
		curInfo.isAccelerated = ((pixelFormatDesc.dwFlags & PFD_GENERIC_FORMAT) == 0);
		curInfo.hasStencil = pixelFormatDesc.cStencilBits >= 8;
		curInfo.hasStereo = ((pixelFormatDesc.dwFlags & PFD_STEREO) != 0);
		curInfo.samples = 0;
		curInfo.depthBits = pixelFormatDesc.cDepthBits;
		curInfo.colorBits = pixelFormatDesc.cColorBits;

		if ( (bestFormat == 0) or
			IsFormatBetter( curInfo, bestInfo, inOptions ) )
		{
			bestFormat = fmt;
			bestInfo = curInfo;
		}
	}
	
	return bestFormat;
}

FFPixelFormat_Win	FFChoosePixelFormat_Win( SBPixelFormatOptions inOptions )
{
	FFPixelFormat_Win	bestFormat = 0;
	StDummyWindowGLContext	dummyContext;
	HDC	glDC = wglGetCurrentDC();
	
	wglChoosePixelFormatARB =
		(PFNWGLCHOOSEPIXELFORMATARBPROC)
		wglGetProcAddress( "wglChoosePixelFormatARB" );
	
	wglGetPixelFormatAttribivARB =
		(PFNWGLGETPIXELFORMATATTRIBIVARBPROC)
		wglGetProcAddress( "wglGetPixelFormatAttribivARB" );
	
	if ( (wglChoosePixelFormatARB != NULL) and (wglGetPixelFormatAttribivARB != NULL) )
	{
		bestFormat = WinChooseFormatNew( glDC, inOptions );
	}
	else
	{
		bestFormat = WinChooseFormatOld( glDC, inOptions );
	}
	return bestFormat;
}
#else // Mac Carbon

#pragma mark -

#include <vector>

#if SB_DEBUG
	#define		LOG_PIXEL_FORMATS	0
#endif

#if LOG_PIXEL_FORMATS
static void LogPixelFormat( FFPixelFormat_Carbon inPixelFormat )
{
	printf("Pixel format %p:\n", inPixelFormat );
	GLint	value;
	if (aglDescribePixelFormat( inPixelFormat, AGL_BUFFER_SIZE, &value ))
	{
		printf("   buffer size %d\n", (int) value );
	}
	else
	{
		printf("   buffer size unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_DOUBLEBUFFER, &value ))
	{
		printf("   double buffer %d\n", (int) value );
	}
	else
	{
		printf("   double buffer unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_STEREO, &value ))
	{
		printf("   stereo buffer %d\n", (int) value );
	}
	else
	{
		printf("   stereo buffer unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_DEPTH_SIZE, &value ))
	{
		printf("   depth buffer %d\n", (int) value );
	}
	else
	{
		printf("   depth buffer unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_STENCIL_SIZE, &value ))
	{
		printf("   stencil buffer %d\n", (int) value );
	}
	else
	{
		printf("   stencil buffer unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_OFFSCREEN, &value ))
	{
		printf("   offscreen %d\n", (int) value );
	}
	else
	{
		printf("   offscreen unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_FULLSCREEN, &value ))
	{
		printf("   full screen %d\n", (int) value );
	}
	else
	{
		printf("   full screen unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_SAMPLE_BUFFERS_ARB, &value ))
	{
		printf("   sample buffers %d\n", (int) value );
	}
	else
	{
		printf("   sample buffers unavailable\n");
	}
	
	if (aglDescribePixelFormat( inPixelFormat, AGL_SAMPLES_ARB, &value ))
	{
		printf("   samples %d\n", (int) value );
	}
	else
	{
		printf("   samples unavailable\n");
	}
	
	fflush( stdout );
}
#else
	#define	LogPixelFormat( x )
#endif

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"

static FFPixelFormat_Carbon CreatePixelFormatForAtts( const std::vector<GLint>& inAtts )
{
	FFPixelFormat_Carbon theFormat = NULL;
#if MAC_OS_X_VERSION_MIN_REQUIRED >= 1050
	theFormat = aglCreatePixelFormat( &inAtts[0] );
#else
	if (aglCreatePixelFormat != 0)
	{
		theFormat = aglCreatePixelFormat( &inAtts[0] );
	}
	else
	{
	#if MAC_OS_X_VERSION_MAX_ALLOWED > 1060
		AGLDevice mainDev = NULL;
	#else
		AGLDevice mainDev = GetMainDevice();
	#endif
		theFormat = aglChoosePixelFormat( &mainDev, 1, &inAtts[0] );
	}
#endif
	return theFormat;
}

FFPixelFormat_Carbon	FFChoosePixelFormat_Carbon( SBPixelFormatOptions inOptions )
{
	FFPixelFormat_Carbon	resultFormat = 0;
	FFPixelFormat_Carbon testFormat = 0;
	
	std::vector<GLint> atts;
	atts.push_back( AGL_DEPTH_SIZE );
	atts.push_back( 16 );		// Keep this at index 1
	atts.push_back( AGL_RGBA );
	atts.push_back( AGL_MINIMUM_POLICY );
	if (aglCreatePixelFormat != 0)
	{
		atts.push_back( AGL_DISPLAY_MASK );
		
		CGDirectDisplayID displayID = CGMainDisplayID();
		CGOpenGLDisplayMask openGLDisplayMask = CGDisplayIDToOpenGLDisplayMask(displayID);
		atts.push_back( openGLDisplayMask );
	}
	
	if ( (inOptions & kSBContextFormat_Window) != 0 )
	{
		atts.push_back( AGL_WINDOW );
		atts.push_back( AGL_DOUBLEBUFFER );
	}
	else
	{
		atts.push_back( AGL_OFFSCREEN );
	}
	
	atts.push_back( AGL_NONE );
	
	testFormat = CreatePixelFormatForAtts( atts );
	
	if (testFormat != 0)
	{
		resultFormat = testFormat;
		LogPixelFormat( testFormat );

		// Try for accelerated?
		if ( (inOptions & kSBContextFormat_PreferAccelerated) != 0 )
		{
			atts.pop_back();
			atts.push_back( AGL_ACCELERATED );
			atts.push_back( AGL_NONE );
			
			testFormat = CreatePixelFormatForAtts( atts );
			
			if (testFormat != 0)
			{
				aglDestroyPixelFormat( resultFormat );
				resultFormat = testFormat;
				LogPixelFormat( testFormat );
			}
			else
			{
				atts.pop_back();
				atts.pop_back();
				atts.push_back( AGL_NONE );
			}
		}
		
		if ( (inOptions & kSBContextFormat_PreferStencil) != 0 )
		{
			atts.pop_back();
			atts.push_back( AGL_STENCIL_SIZE );
			atts.push_back( 8 );
			atts.push_back( AGL_NONE );
			
			testFormat = CreatePixelFormatForAtts( atts );
			
			if (testFormat != 0)
			{
				aglDestroyPixelFormat( resultFormat );
				resultFormat = testFormat;
				LogPixelFormat( testFormat );
			}
			else
			{
				atts.pop_back();
				atts.pop_back();
				atts.pop_back();
				atts.push_back( AGL_NONE );
			}
		}
		
		if ( (inOptions & kSBContextFormat_Stereo) != 0 )
		{
			atts.pop_back();
			atts.push_back( AGL_STEREO );
			atts.push_back( AGL_NONE );

			testFormat = CreatePixelFormatForAtts( atts );
			
			if (testFormat != 0)
			{
				aglDestroyPixelFormat( resultFormat );
				resultFormat = testFormat;
				LogPixelFormat( testFormat );
			}
			else
			{
				atts.pop_back();
				atts.pop_back();
				atts.push_back( AGL_NONE );
			}
		}
		
		if ( (inOptions & kSBContextFormat_PreferMultisample) != 0 )
		{
			atts.pop_back();
			atts.push_back( AGL_MULTISAMPLE );
			atts.push_back( AGL_SAMPLE_BUFFERS_ARB );
			atts.push_back( 1 );
			atts.push_back( AGL_SAMPLES_ARB );
			atts.push_back( 2 );
			atts.push_back( AGL_NONE );
			
			testFormat = CreatePixelFormatForAtts( atts );
			
			if (testFormat == 0)
			{
				atts.pop_back();
				atts.pop_back();
				atts.push_back( 4 );
				atts.push_back( AGL_NONE );
				
				testFormat = CreatePixelFormatForAtts( atts );
			}
			
			if (testFormat != 0)
			{
				aglDestroyPixelFormat( resultFormat );
				resultFormat = testFormat;
				LogPixelFormat( testFormat );
			}
		}
		
		// Try for 24 depth bits if possible
		GLint	depthBits;
		if ( aglDescribePixelFormat( resultFormat, AGL_DEPTH_SIZE, &depthBits )
			and (depthBits < 24) )
		{
			atts[1] = 24;
			
			testFormat = CreatePixelFormatForAtts( atts );
			
			if (testFormat != 0)
			{
				aglDestroyPixelFormat( resultFormat );
				resultFormat = testFormat;
			}
		}
	}
	
	return resultFormat;
}

#endif

#pragma mark -

/*!
	@function	SBChoosePixelFormat
	
	@abstract	Look for the best pixel format meeting some requirements.
	
	@discussion	We only consider formats that support OpenGL and RGB pixels,
				with a depth buffer at least 16 bits deep.
				
				We require a format that supports windowed rendering if the
				kSBContextFormat_Window flag is set, and require a format that
				supports rendering to a bitmap if that flag is clear.
				
				When kSBContextFormat_Window is set, we also require double
				buffering.
				
				The preferences are in this order of priority:
				
				* Accelerated if kSBContextFormat_PreferAccelerated is set,
				  software rendered if kSBContextFormat_PreferAccelerated is
				  clear.
				
				* 8 bits or more of stencil if kSBContextFormat_PreferStencil
				  is set.
				
				* Multisample capability if kSBContextFormat_PreferMultisample
				  is set, the more samples, the better.
				
				* The more depth bits, the better.
				
				* The more color bits, the better.
	
	@param		inOptions		Some of the ESBContextFormatOptions values ored
								together.
	
	@result		A pixel format, or 0 on failure.
*/
#if QUESA_OS_COCOA
FFPixelFormat_Cocoa	FFChoosePixelFormat_Cocoa( SBPixelFormatOptions inOptions )
{
	FFPixelFormat_Cocoa bestFormat = CocoaChoosePixelFormat( inOptions );
	return bestFormat;
}
#endif


/*!
	@function	SBDisposePixelFormat
	
	@abstract	Release a pixel format that is no longer needed.
	
	@param		inFormat		A pixel format.
*/
#if WIN32
void			FFDisposePixelFormat_Win( FFPixelFormat_Win inFormat )
{
	#pragma unused( inFormat )
}
#else
void			FFDisposePixelFormat_Mac( FFPixelFormat_Carbon inFormat )
{
	aglDestroyPixelFormat( inFormat );
}

	#if QUESA_OS_COCOA
	void		FFDisposePixelFormat_Mac( FFPixelFormat_Cocoa inFormat )
	{
		CocoaDisposePixelFormat( inFormat );
	}
	#endif
#endif

